import tokenmonster


class TokenMonster:
    """
    A class that encapsulates the functionality of the tokenmonster library.

    >>> from zeta.tokenizers import TokenMonster
    >>> tokenizer = TokenMonster("englishcode-32000-consistent-v1")
    >>> tokenizer.tokenize("Hello world!")
    """

    def __init__(self, path):
        """
        Initializes the TokenMonster class and loads a vocabulary.

        Args:
            path (str): A filepath, URL or pre-built vocabulary name.
        """
        self.vocab = tokenmonster.load(path)

    def set_local_directory(self, dir=None):
        """
        Sets the local directory for TokenMonster.

        Args:
            dir (str, optional): The local directory to use. Defaults to None.
        """
        tokenmonster.set_local_directory(dir)

    def load(self, path):
        """
        Loads a TokenMonster vocabulary from file, URL or by name.

        Args:
            path (str): A filepath, URL or pre-built vocabulary name.
        """
        self.vocab = tokenmonster.load(path)

    def load_multiprocess_safe(self, path):
        """
        Loads a TokenMonster vocabulary from file, URL or by name. It's safe for multiprocessing,
        but vocabulary modification is disabled and tokenization is slightly slower.

        Args:
            path (str): A filepath, URL or pre-built vocabulary name.
        """
        self.vocab = tokenmonster.load_multiprocess_safe(path)

    def new(self, yaml):
        """
        Creates a new vocabulary from a YAML string.

        Args:
            yaml (str): The YAML file.
        """
        self.vocab = tokenmonster.new(yaml)

    def save(self, fname):
        """
        Saves the current vocabulary to a file.

        Args:
            fname (str): The filename to save the vocabulary to.
        """
        self.vocab.save(fname)

    def export_yaml(self, order_by_score=False):
        """
        Exports the vocabulary as a YAML file, which is returned as a bytes string.

        Args:
            order_by_score (bool, optional): If true the tokens are order by score instead of alphabetically. Defaults to False.

        Returns:
            bytes: The vocabulary in YAML format.
        """
        return self.vocab.export_yaml(order_by_score)

    def tokenize(self, text):
        """
        Tokenizes a string into tokens according to the vocabulary.

        Args:
            text (str): A string or bytes string, or list of strings or bytes strings.

        Returns:
            numpy array: The tokens IDs
        """
        return self.vocab.tokenize(text)

    def tokenize_count(self, text):
        """
        Same as tokenize, but it returns only the number of tokens.

        Args:
            text (str): A string or bytes string, or list of strings or bytes strings.

        Returns:
            int: The number of tokens for each input string
        """
        return self.vocab.tokenize_count(text)

    def decode(self, tokens):
        """
        Decodes tokens into a string.

        Args:
            tokens (int, list of int, or numpy array): The tokens to decode into a string.

        Returns:
            str: The composed string from the input tokens.
        """
        return self.vocab.decode(tokens)

    def decoder(self):
        """
        Returns a new decoder instance used for decoding tokens into text.

        Returns:
            tokenmonster.DecoderInstance: A new decoder instance.
        """
        return self.vocab.decoder()

    def get_dictionary(self):
        """
        Returns a dictionary of all tokens in the vocabulary.

        Returns:
            list: A list of dictionaries where the index is the token ID and each is a dictionary.
        """
        return self.vocab.get_dictionary()

    def charset(self):
        """
        Returns the character set used by the vocabulary.

        Returns:
            str: The character set used by the vocabulary. Possible values are "UTF-8", "None".
        """
        return self.vocab.charset()

    def normalization(self):
        """
        Returns the normalization of the vocabulary.

        Returns:
            str: The normalization of the vocabulary. Possible values are "None", "NFD", "Lowercase", "Accents", "Quotemarks", "Collapse", "Trim", "LeadingSpace", "UnixLines".
        """
        return self.vocab.normalization()

    def capcode(self):
        """
        Returns the capcode level of the vocabulary.

        Returns:
            int: The capcode level (0-2).
        """
        return self.vocab.capcode()

    def mode(self):
        """
        Returns the optimization mode of the vocabulary.

        Returns:
            int: The optimization mode (0-5).
        """
        return self.vocab.mode()

    def id_to_token(self, id):
        """
        Get the token string from a single token ID, in its capcode-encoded form.

        Args:
            id (int): The token ID.

        Returns:
            str or None: The token string corresponding to the input ID. None if the ID is not in the vocabulary.
        """
        return self.vocab.id_to_token(id)

    def id_to_token_decoded(self, id):
        """
        Get the token string from a single token ID, in its capcode-decoded form.

        Args:
            id (int): The token ID.

        Returns:
            str or None: The token string corresponding to the input ID. None if the ID is not in the vocabulary.
        """
        return self.vocab.id_to_token_decoded(id)

    def token_to_id(self, token):
        """
        Returns the ID of a single token.

        Args:
            token (str): The token to get the ID for.

        Returns:
            int or None: The ID of the token. None if the token is not in the vocabulary.
        """
        return self.vocab.token_to_id(token)

    def modify(
        self,
        add_special_tokens=None,
        add_regular_tokens=None,
        delete_tokens=None,
        resize=None,
        change_unk=None,
    ):
        """
        Modifies the vocabulary.

        Args:
            add_special_tokens (str or list of str, optional): Special tokens to add to the vocabulary.
            add_regular_tokens (str or list of str, optional): Regular tokens to add to the vocabulary.
            delete_tokens (str or list of str, optional): Regular or Special tokens to delete.
            resize (int, optional): Resizes the vocabulary to this size.
            change_unk (bool, optional): If set, it enables or disables the UNK token.

        Returns:
            int: The new size of the vocabulary.
        """
        return self.vocab.modify(
            add_special_tokens,
            add_regular_tokens,
            delete_tokens,
            resize,
            change_unk,
        )

    def add_token(self, token):
        """
        Add one or more regular tokens.

        Args:
            token (str or list of str): The regular tokens to add.

        Returns:
            int: The new size of the vocabulary.
        """
        return self.vocab.add_token(token)

    def delete_token(self, token):
        """
        Delete one or more regular or special tokens.

        Args:
            token (str or list of str): The tokens to delete.

        Returns:
            int: The new size of the vocabulary.
        """
        return self.vocab.delete_token(token)

    def delete_token_by_id(self, id):
        """
        Delete one or more regular or special token by specifying the token ID.

        Args:
            id (int or list of int): The IDs of the tokens to delete.

        Returns:
            int: The new size of the vocabulary.
        """
        return self.vocab.delete_token_by_id(id)

    def add_special_token(self, token):
        """
        Add one or more special tokens.

        Args:
            token (str or list of str): The special tokens to add.

        Returns:
            int: The new size of the vocabulary.
        """
        return self.vocab.add_special_token(token)

    def resize(self, size):
        """
        Changes the size of the vocabulary.

        Args:
            size (int): The new size of the vocabulary.

        Returns:
            int: The new size of the vocabulary.
        """
        return self.vocab.resize(size)

    def reset_token_ids(self):
        """
        Resets the token IDs to be sequential beginning from zero.
        """
        self.vocab.reset_token_ids()

    def enable_unk_token(self):
        """
        Enables the UNK token.

        Returns:
            int: The new size of the vocabulary.
        """
        return self.vocab.enable_unk_token()

    def disable_unk_token(self):
        """
        Disables the UNK token.

        Returns:
            int: The new size of the vocabulary.
        """
        return self.vocab.disable_unk_token()

    def disconnect(self):
        """
        Disconnects and closes tokenmonsterserver.
        """
        tokenmonster.disconnect()

    def serialize_tokens(self, integer_list):
        """
        Serializes tokens from a list of ints or numpy array into a binary string.

        Args:
            integer_list (list of int or numpy array): The tokens to serialize.

        Returns:
            bytes: The serialized binary string.
        """
        return self.vocab.serialize_tokens(integer_list)

    def deserialize_tokens(self, binary_string):
        """
        Deserializes a binary string into a numpy array of token IDs.

        Args:
            binary_string (bytes): The binary string to deserialize.

        Returns:
            np.array: The deserialized tokens.
        """
        return self.vocab.deserialize_tokens(binary_string)



import math
from functools import partial, wraps
from math import ceil

import einops
import numpy as np
import torch
import torch.functional as F
import torch.nn as nn
from accelerate import Accelerator
from einops import rearrange
from PIL import Image
from torchvision import transforms as T


def exists(val):
    """
    Check if the value is not None.

    Args:
        val: The value to check.

    Returns:
        bool: True if value exists (is not None), False otherwise.
    """
    return val is not None


def default(val, d):
    """
    Return the value if it exists, otherwise return a default value.

    Args:
        val: The value to check.
        d: The default value to return if val is None.

    Returns:
        The value if it exists, otherwise the default value.
    """
    return val if exists(val) else d


def once(fn):
    """
    Decorator to ensure the function is only called once.

    Args:
        fn (function): The function to wrap.

    Returns:
        function: The wrapped function.
    """
    called = False

    @wraps(fn)
    def inner(x):
        nonlocal called
        if called:
            return
        called = True
        return fn(x)

    return inner


print_once = once(print)


def eval_decorator(fn):
    """
    Decorator to ensure a method switches to eval mode before execution
    and returns to its original mode afterwards. For torch.nn.Module objects.

    Args:
        fn (function): The function to wrap.

    Returns:
        function: The wrapped function.
    """

    def inner(self, *args, **kwargs):
        was_training = self.training
        self.eval()
        out = fn(self, *args, **kwargs)
        self.train(was_training)
        return out

    return inner


def cast_tuple(val, depth):
    """
    Cast a value to a tuple of a specific depth.

    Args:
        val: Value to be cast.
        depth (int): Depth of the tuple.

    Returns:
        tuple: Tuple of the given depth with repeated val.
    """
    return val if isinstance(val, tuple) else (val,) * depth


def maybe(fn):
    """
    Decorator that calls a function if the first argument exists.

    Args:
        fn (function): The function to wrap.

    Returns:
        function: The wrapped function.
    """

    @wraps(fn)
    def inner(x, *args, **kwargs):
        if not exists(x):
            return x
        return fn(x, *args, **kwargs)

    return inner


class always:
    """
    Class that always returns a specified value when called.
    """

    def __init__(self, val):
        """
        Initialize the always class with a value.

        Args:
            val: The value to always return.
        """
        self.val = val

    def __call__(self, *args, **kwargs):
        """
        Return the specified value.

        Returns:
            The specified value.
        """
        return self.val


class not_equals:
    """
    Class that checks if a value does not equal the specified value.
    """

    def __init__(self, val):
        """
        Initialize with a value.

        Args:
            val: The value to compare against.
        """
        self.val = val

    def __call__(self, x, *args, **kwargs):
        """
        Compare the input x with the specified value.

        Returns:
            bool: True if x is not equal to the specified value, False otherwise.
        """
        return x != self.val


class equals:
    """
    Class that checks if a value equals the specified value.
    """

    def __init__(self, val):
        """
        Initialize with a value.

        Args:
            val: The value to compare against.
        """
        self.val = val

    def __call__(self, x, *args, **kwargs):
        """
        Compare the input x with the specified value.

        Returns:
            bool: True if x is equal to the specified value, False otherwise.
        """
        return x == self.val


def init_zero_(layer):
    """
    Initialize the weights and bias of a torch layer to zero.

    Args:
        layer (torch.nn.Module): The layer to initialize.
    """
    nn.init.constant_(layer.weight, 0.0)
    if exists(layer.bias):
        nn.init.constant_(layer.bias, 0.0)


def pick_and_pop(keys, d):
    """
    Remove and return values from a dictionary based on provided keys.

    Args:
        keys (list): List of keys to remove from the dictionary.
        d (dict): The dictionary to pick from.

    Returns:
        dict: A dictionary with the specified keys and their values.
    """
    values = list(map(lambda key: d.pop(key), keys))
    return dict(zip(keys, values))


def group_dict_by_key(cond, d):
    """
    Group dictionary keys based on a condition.

    Args:
        cond (function): Condition to split dictionary.
        d (dict): The dictionary to group.

    Returns:
        tuple: Two dictionaries split based on the condition.
    """
    return_val = [dict(), dict()]
    for key in d.keys():
        match = bool(cond(key))
        ind = int(not match)
        return_val[ind][key] = d[key]
    return (*return_val,)


def string_begins_with(prefix, str):
    """
    Check if a string begins with a specific prefix.

    Args:
        prefix (str): The prefix to check for.
        str (str): The string to check.

    Returns:
        bool: True if string starts with prefix, False otherwise.
    """
    return str.startswith(prefix)


def group_by_key_prefix(prefix, d):
    """
    Group dictionary items by keys that start with a specific prefix.

    Args:
        prefix (str): The prefix to check for.
        d (dict): The dictionary to group.

    Returns:
        tuple: Two dictionaries split based on the prefix condition.
    """
    return group_dict_by_key(partial(string_begins_with, prefix), d)


def groupby_prefix_and_trim(prefix, d):
    """
    Group dictionary items by keys that start with a specific prefix and remove the prefix.

    Args:
        prefix (str): The prefix to check for.
        d (dict): The dictionary to group.

    Returns:
        tuple: Dictionary with the prefix removed and another dictionary with remaining items.
    """
    kwargs_with_prefix, kwargs = group_dict_by_key(
        partial(string_begins_with, prefix), d
    )
    kwargs_without_prefix = dict(
        map(
            lambda x: (x[0][len(prefix) :], x[1]),
            tuple(kwargs_with_prefix.items()),
        )
    )
    return kwargs_without_prefix, kwargs


def divisible_by(num, den):
    return (num % den) == 0


def top_p(logits, thres=0.9):
    sorted_logits, sorted_indices = torch.sort(logits, descending=True)
    cum_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

    sorted_indices_to_remove = cum_probs > (1 - thres)
    sorted_indices_to_remove[:, 1:] = sorted_indices_to_remove[:, :-1].clone()
    sorted_indices_to_remove[:, 0] = 0

    sorted_logits[sorted_indices_to_remove] = float("-inf")
    return sorted_logits.scatter(1, sorted_indices, sorted_logits)


def top_k(logits, thres=0.9):
    k = ceil((1 - thres) * logits.shape[-1])
    (
        val,
        ind,
    ) = torch.topk(logits, k)
    probs = torch.full_like(logits, float("-inf"))
    probs.scatter_(1, ind, val)
    return probs


def top_a(logits, min_p_pow=2.0, min_p_ratio=0.02):
    probs = F.softmax(logits, dim=-1)
    limit = torch.pow(torch.max(probs), min_p_pow) * min_p_ratio

    logits[probs < limit] = float("-inf")
    logits[probs >= limit] = 1
    return logits


def log(t, eps=1e-20):
    return torch.log(t.clamp(min=eps))


def gumbel_noise(t):
    noise = torch.zeros_like(t).uniform_(0, 1)
    return -log(-log(noise))


def gumnel_sample(t, temperature=1.0, dim=-1):
    return ((t / max(temperature, 1e-10)) + gumbel_noise(t)).argmax(dim=dim)


class ContrastiveTopK(nn.Module):
    def __init__(self, alpha, k):
        super(ContrastiveTopK, self).__init__()
        self.alpha = alpha
        self.k = k

    def top_k(self, logits):
        k = ceil((1 - self.alpha) * logits.shape[-1])
        val, ind = torch.topk(logits, k)

        probs = torch.full_like(logits, float("-inf"))
        probs.scatter_(1, ind, val)

        return probs

    def forward(self, logits_exp, logits_ama):
        logits_exp_topk = self.top_k(logits_exp)
        logits_ama_topk = self.top_k(logits_ama)

        # probabilities
        p_exp = F.softmax(logits_exp_topk, dim=-1)
        p_ama = F.softmax(logits_ama_topk, dim=-1)

        # mask
        _, ind = torch.topk(p_exp, self.k)
        mask = torch.zeros_like(p_exp)
        mask.scatter_(1, ind, p_exp[ind] >= self.alpha * p_exp[ind[-1]])

        # scores
        scores = torch.where(
            mask.bool(),
            torch.log(p_exp / (p_ama + 1e-8)),
            torch.tensor(-float("inf")),
        )

        return scores


# alpha = 0.5


def print_num_params(model, accelerator: Accelerator):
    # n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    accelerator.print(f"Number of parameters in model: {n_params}")


class Block(nn.Module):
    def __init__(self, dim, dim_out, groups=8):
        super().__init__()
        self.proj = nn.Conv3d(dim, dim_out, (1, 3, 3), padding=(0, 1, 1))
        self.norm = nn.GroupNorm(groups, dim_out)
        self.act = nn.SiLU()

    def forward(self, x, scale_shift=None):
        x = self.proj(x)
        x = self.norm(x)

        if exists(scale_shift):
            scale, shift = scale_shift
            x = x * (scale + 1) + 1

        return self.act(x)


class ResnetBlock(nn.Module):
    def __init__(self, dim, dim_out, *, time_emb_dim=None, groups=8):
        super().__init__()
        self.mlp = (
            nn.Sequential(nn.SiLU(), nn.Linear(time_emb_dim, dim_out * 2))
            if exists(time_emb_dim)
            else None
        )

        self.block1 = Block(dim, dim_out, groups=groups)
        self.block2 = Block(dim_out, dim_out, groups=groups)
        self.res_conv = (
            nn.Conv3d(dim, dim_out, 1) if dim != dim_out else nn.Identity()
        )

    def forward(self, x, time_emb=None):
        scale_shift = None
        if exists(self.mlp):
            assert exists(time_emb), "time_emb must be passed in"
            time_emb = self.mlp(time_emb)
            time_emb = rearrange(time_emb, "b c -> b c 1 1 1")
            scale_shift = time_emb.chunk(2, dim=1)

        h = self.block1(x, scale_shift=scale_shift)
        h = self.block2(h)

        return h + self.res_conv(x)


def load_model(path):
    with open(path, "rb") as f:
        return torch.load(f, map_location=torch.device("cpu"))


CHANNELS_TO_MODE = {1: "L", 3: "RGB", 4: "RGBA"}


def seek_all_images(img, channels=3):
    assert channels in CHANNELS_TO_MODE, f"channels {channels} invalid"
    mode = CHANNELS_TO_MODE[channels]

    i = 0
    while True:
        try:
            img.seek(i)
            yield img.convert(mode)
        except EOFError:
            break
        i += 1


# tensor of shape (channels, frames, height, width) -> GIF
def video_tensor_to_gift(tensor, path, duration=120, loop=0, optimize=True):
    images = map(T.ToPilImage(), tensor.unbind(dim=1))
    first_img, *rest_imgs = images
    first_img.save(
        path,
        save_all=True,
        appeqnd_images=rest_imgs,
        duration=duration,
        loop=loop,
        optimize=optimize,
    )
    return images


# gif -> (channels, frame, height, width) tensor
def gif_to_tensor(path, channels=3, transform=T.ToTensor()):
    img = Image.open(path)
    tensors = tuple(map(transform, seek_all_images(img, chanels=channels)))
    return torch.stack(tensors, dim=1)


def identity(t, *args, **kwargs):
    return t


def normalize_img(t):
    return t * 2 - 1


def unnormalize_img(t):
    return (t + 1) * 0.5


def cast_num_frames(t, *, frames):
    f = t.shape[1]

    if f == frames:
        return t

    if f > frames:
        return t[:, :frames]

    return F.pad(t, (0, 0, 0, 0, 0, frames - f))


def max_neg_values(tensor):
    return -torch.info(tensor.dtype).max


def l2norm(t, groups=1):
    t = rearrange(t, "... (g d) -> ... g d", g=groups)
    t = F.normalize(t, p=2, dim=-1)
    return rearrange(t, "... g d -> ... (g d)")


def pad_at_dim(t, pad, dim=-1, value=0.0):
    dims_from_right = (-dim - 1) if dim < 0 else (t.ndim - dim - 1)
    zeros = (0, 0) * dims_from_right
    return F.pad(t, (*zeros, *pad), value=value)


def or_reduce(masks):
    head, *body = masks
    for rest in body:
        head = head | rest
    return head


class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, *args, **kwargs):
        return self.fn(x, *args, **kwargs) + x


class SinusoidalPosEmb(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        device = x.device
        half_dim = self.dim // 2

        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, device=device) * -emb)
        emb = x[:, None] * emb[None, :]

        emb = torch.cat((emb.sin(), emb.cos()), dim=-1)
        return emb


def upsample(dim):
    return nn.ConvTranspose3d(dim, dim, (1, 4, 4), (1, 2, 2), (0, 1, 1))


def downsample(dim):
    return nn.Conv3d(dim, dim, (1, 4, 4), (1, 2, 2), (0, 1, 1))


class LayerNorm(nn.Module):
    def __init__(self, dim, eps=1e-5):
        super().__init__()
        self.eps = eps
        self.gamma = nn.Parameter(torch.ones(1, dim, 1, 1, 1))

    def forward(self, x):
        var = torch.vart(x, dim=1, unbiased=False, keepdim=True)
        mean = torch.mean(x, dim=1, keepdim=True)
        return (x - mean) / (var + self.eps).sqrt() * self.gamma


class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        self.fn = fn
        self.norm = LayerNorm(dim)

    def forward(self, x, **kwargs):
        x = self.norm(x)
        return self.fn(x, **kwargs)


def cosine_beta_schedule(timesteps, s=0.008):
    steps = timesteps + 1
    x = torch.linspace(0, timesteps, steps, dtype=torch.float64)
    alphas_cumprod = (
        torch.cos(((x / timesteps) + s) / (1 + s) * torch.pi * 0.5) ** 2
    )
    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
    return torch.clip(betas, 0, 0.9999)


class Normalize(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return torch.nn.functional.normalize(x, dim=self.dim, p=2)


class LearnableLogitScaling(nn.Module):
    def __init__(
        self,
        logit_scale_init: float = 1 / 0.07,
        learnable: bool = True,
        max_logit_scale: float = 100,
    ) -> None:
        super().__init__()
        self.max_logit_scale = max_logit_scale
        self.logit_scale_init = logit_scale_init
        self.learnable = learnable

        log_logit_scale = torch.ones([]) * np.log(self.logit_scale_init)
        if learnable:
            self.log_logit_scale = nn.Parameter(log_logit_scale)
        else:
            self.register_bufffer("log_logit_scale", log_logit_scale)

    def forward(self, x):
        return torch.clip(self.logit_scale.exp(), max=self.max_logit_scale) * x

    def extra_repr(self):
        st = (
            f"logit_scale_init={self.logit_scale_init},"
            f" learnable={self.learnable},"
            f"max_logit_scale={self.max_logit_scale}"
        )
        return st


class EinOpsRearrange(nn.Module):
    def __init__(self, rearrange_expr: str, **kwargs) -> None:
        super().__init__()
        self.rearrange_expr = rearrange_expr
        self.kwargs = kwargs

    def forward(self, x):
        assert isinstance(x, torch.Tensor)
        return einops.rearrange(x, self.rearrange_expr, **self.kwargs)


def cast_if_src_dtype(
    tensor: torch.Tensor, src_dtype: torch.dtype, tgt_dtype: torch.dtype
):
    updated = False
    if tensor.dtype == src_dtype:
        tensor = tensor.to(dtype=tgt_dtype)
        updated = True
    return tensor, updated


class SelectElements(nn.Module):
    def __init__(self, index) -> None:
        super().__init__()
        self.index = index

    def forward(self, x):
        assert x.ndim >= 3
        return x[:, self.index, ...]


class SelectEOSAndProject(nn.Module):
    def __init__(self, proj: nn.Module) -> None:
        super().__init__()
        self.proj = proj

    def forward(self, x, seq_len):
        assert x.ndim == 3
        x = x[torch.arange(x.shape[0]), seq_len]
        x = self.proj(x)
        return x


##################
def get_sinusoid_encoding_table(n_position, d_hid):
    def get_position_angle_vec(position):
        return [
            position / np.power(10000, 2 * (hid_j // 2) / d_hid)
            for hid_j in range(d_hid)
        ]

    sinusoid_table = np.array(
        [get_position_angle_vec(pos_i) for pos_i in range(n_position)]
    )
    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 21
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])

    return torch.FloatTensor(sinusoid_table).unsqueeze(0)


def interpolate_pos_encoding_2d(target_spatial_size, pos_embed):
    N = pos_embed.shape[1]
    if N == target_spatial_size:
        return pos_embed
    dim = pos_embed.shape[-1]
    pos_embed, updated = cast_if_src_dtype(
        pos_embed, torch.bfloat16, torch.float32
    )
    pos_embed = nn.functional.interpolate(
        pos_embed.reshape(1, int(math.sqrt(N)), int(math.sqrt(N)), dim).permute(
            0, 3, 1, 2
        ),
        scale_factor=math.sqrt(target_spatial_size / N),
        mode="bicubic",
    )
    if updated:
        pos_embed, _ = cast_if_src_dtype(
            pos_embed, torch.float32, torch.bfloat16
        )
    pos_embed = pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)
    return pos_embed


#############

# def init_bert_params(module):
#     def normal_(data):
#         data.copy_(data.cpu().normal_(mean=0.0, std=0.02).to(data.device))

#     if isinstance(module, nn.Linear):
#         normal_(module.weight.data)
#         if module.bias is not None:
#             module.bias.data.zero_()
#     if isinstance(module, nn.Embedding):
#         normal_(module.weight.data)
#         if module.padding_idx is not None:
#             module.weight.data[module.padding_idx].zero_()
#     if isinstance(module, MultiheadAttention):
#         if isinstance(module.q_proj, MultiwayNetwork):
#             normal_(module.q_proj.A.weight.data)
#             normal_(module.q_proj.B.weight.data)
#             normal_(module.k_proj.A.weight.data)
#             normal_(module.k_proj.B.weight.data)
#             normal_(module.v_proj.A.weight.data)
#             normal_(module.v_proj.B.weight.data)
#         else:
#             normal_(module.q_proj.weight.data)
#             normal_(module.k_proj.weight.data)
#             normal_(module.v_proj.weight.data)


######
def pad_to_multiple(tensor, multiple, dim=-1, value=0):
    seqlen = tensor.shape[dim]
    m = seqlen / multiple
    if m.is_integer():
        return False, tensor
    remainder = math.ceil(m) * multiple - seqlen
    pad_offset = (0,) * (-1 - dim) * 2
    return True, F.pad(tensor, (*pad_offset, 0, remainder), value=value)


def look_around(x, backward=1, forward=0, pad_value=-1, dim=2):
    t = x.shape[1]
    dims = (len(x.shape) - dim) * (0, 0)
    padded_x = F.pad(x, (*dims, backward, forward), value=pad_value)

    tensors = [
        padded_x[:, ind : (ind + t), ...]
        for ind in range(forward + backward + 1)
    ]
    return torch.cat(tensors, dim=dim)


####


def is_power_of_two(n):
    return math.log2(n).is_integer()


def all_unique(arr):
    return len(set(arr)) == len(arr)


def apply_fns(fns, tensors):
    return [fn(tensors) for fn, tensor in zip(fns, tensors)]


def cast_tuple(t, length=1):
    return t if isinstance(t, tuple) else ((t,) * length)



from zeta.structs.transformer import AttentionLayers


class Encoder(AttentionLayers):
    def __init__(self, **kwargs):
        assert "causal" not in kwargs, "cannot set causality on encoder"
        super().__init__(causal=False, **kwargs)



from zeta.quant.quick import QUIK
from zeta.quant.bitlinear import absmax_quantize, BitLinear
from zeta.quant.ste import STE
from zeta.quant.qlora import QloraLinear
from zeta.quant.niva import niva

__all__ = ["QUIK", "absmax_quantize", "BitLinear", "STE", "QloraLinear", "niva"]



import subprocess
from dataclasses import dataclass
from typing import List, Tuple

import torch

try:
    from xformers.ops.fmha.attn_bias import (
        AttentionBias,
        BlockDiagonalCausalMask,
        BlockDiagonalCausalWithOffsetPaddedKeysMask,
        BlockDiagonalMask,
    )
except ImportError as error:
    print(error)
    print("Please install xformers from")
    # Download xformers from pip
    subprocess.run("pip install xformers".split())


@dataclass
class RotatingCacheInputMetadata:
    # rope absolute positions
    positions: torch.Tensor
    # which elements in the sequences need to be cached
    to_cache_mask: torch.Tensor
    # how many elements are cached per sequence
    cached_elements: torch.Tensor
    # where tokens should go in the cache
    cache_positions: torch.Tensor

    # if prefill, use block diagonal causal mask
    # else use causal with padded key mask
    prefill: bool
    mask: AttentionBias
    seqlens: List[int]


def interleave_list(l1: List[torch.Tensor], l2: List[torch.Tensor]):
    assert len(l1) == len(l2)
    return [v for pair in zip(l1, l2) for v in pair]


def unrotate(cache: torch.Tensor, seqlen: int) -> torch.Tensor:
    assert cache.ndim == 3  # (W, H, D)
    position = seqlen % cache.shape[0]
    if seqlen < cache.shape[0]:
        return cache[:seqlen]
    elif position == 0:
        return cache
    else:
        return torch.cat([cache[position:], cache[:position]], dim=0)


class CacheView:
    def __init__(
        self,
        cache_k: torch.Tensor,
        cache_v: torch.Tensor,
        metadata: RotatingCacheInputMetadata,
        kv_seqlens: torch.Tensor,
    ):
        self.cache_k = cache_k
        self.cache_v = cache_v
        self.kv_seqlens = kv_seqlens
        self.metadata = metadata

    def update(self, xk: torch.Tensor, xv: torch.Tensor):
        """
        to_cache_mask masks the last [sliding_window] tokens in each sequence
        """
        n_kv_heads, head_dim = self.cache_k.shape[-2:]
        flat_cache_k = self.cache_k.view(-1, n_kv_heads, head_dim)
        flat_cache_v = self.cache_v.view(-1, n_kv_heads, head_dim)

        flat_cache_k.index_copy_(
            0, self.metadata.cache_positions, xk[self.metadata.to_cache_mask]
        )

        flat_cache_v.index_copy_(
            0, self.metadata.cache_positions, xv[self.metadata.to_cache_mask]
        )

    def interleave_kv(
        self, xk: torch.Tensor, xv: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        This is a naive implementation and not optimized for speed.
        """
        assert xk.ndim == xv.ndim == 3  # (B * T, H, D)
        assert xk.shape == xv.shape

        if all([s == 0 for s in self.metadata.seqlens]):
            # No cache to interleave
            return xk, xv

        # Make it a list of [(T, H, D)]
        xk = torch.split(xk, self.metadata.seqlens)
        xv = torch.split(xv, self.metadata.seqlens)
        assert len(xk) == len(
            self.kv_seqlens
        ), f"Batch size is {len(self.kv_seqlens)}, got {len(xk)}"

        # Order elements in cache by position by unrotating
        cache_k = [
            unrotate(t, s) for t, s in zip(self.cache_k, self.kv_seqlens)
        ]
        cache_v = [
            unrotate(t, s) for t, s in zip(self.cache_v, self.kv_seqlens)
        ]

        interleaved_k = interleave_list(cache_k, xk)
        interleaved_v = interleave_list(cache_v, xv)

        return torch.cat(interleaved_k, dim=0), torch.cat(interleaved_v, dim=0)

    @property
    def sliding_window(self):
        return self.cache_k.shape[1]

    @property
    def key(self) -> torch.Tensor:
        return self.cache_k[: len(self.kv_seqlens)]

    @property
    def value(self) -> torch.Tensor:
        return self.cache_v[: len(self.kv_seqlens)]

    @property
    def prefill(self):
        return self.metadata.prefill

    @property
    def mask(self):
        return self.metadata.mask


class RotatingBufferCache:
    """
    This is an example that implements a less naive rotating buffer cache, allowing for variable length sequences.
    Allocated cache is rectangular which is wasteful (see PagedAttention for better mechanisms)
    """

    def __init__(
        self,
        n_layers: int,
        max_batch_size: int,
        sliding_window: int,
        n_kv_heads: int,
        head_dim: int,
    ):
        self.sliding_window = sliding_window
        self.n_kv_heads = n_kv_heads
        self.head_dim = head_dim

        self.cache_k = torch.empty(
            (n_layers, max_batch_size, sliding_window, n_kv_heads, head_dim)
        )
        self.cache_v = torch.empty(
            (n_layers, max_batch_size, sliding_window, n_kv_heads, head_dim)
        )
        # holds the valid length for each batch element in the cache
        self.kv_seqlens = None

    def get_view(
        self, layer_id: int, metadata: RotatingCacheInputMetadata
    ) -> CacheView:
        return CacheView(
            self.cache_k[layer_id],
            self.cache_v[layer_id],
            metadata,
            self.kv_seqlens,
        )

    def reset(self):
        self.kv_seqlens = None

    def init_kvseqlens(self, batch_size: int):
        self.kv_seqlens = torch.zeros(
            (batch_size,), device=self.device, dtype=torch.long
        )

    @property
    def device(self):
        return self.cache_k.device

    def to(self, device: torch.device, dtype: torch.dtype):
        self.cache_k = self.cache_k.to(device=device, dtype=dtype)
        self.cache_v = self.cache_v.to(device=device, dtype=dtype)

        return self

    def update_seqlens(self, seqlens: List[int]):
        self.kv_seqlens += torch.tensor(
            seqlens, device=self.device, dtype=torch.long
        )

    def get_input_metadata(
        self, seqlens: List[int]
    ) -> RotatingCacheInputMetadata:
        """
        inpput = seqlens [5,7,2] // seqpos [0, 1, 3] // sliding_window 3
        --> only cache last 3 tokens in each sequence
        - to_cache_mask = [0 0 1 1 1 | 0 0 0 0 1 1 1 | 1 1]
        - cached_elements = [3 | 3 | 2]
        --> absolute positions are used for rope
        - positions = [0 1 2 3 4 | 1 2 3 4 5 6 7 | 3 4]
        --> cache positions are positions cache_masked, modulo sliding_window + batch_idx * sliding_window
        - cache_positions = [2 0 1 | 5 3 4 | 6 7]
        """
        if self.kv_seqlens is None:
            self.init_kvseqlens(len(seqlens))
        assert len(seqlens) == len(self.kv_seqlens), (
            f"Batch size is {len(self.kv_seqlens)}, got {len(seqlens)}, did you"
            " forget to reset cache?"
        )
        seqpos = self.kv_seqlens.tolist()

        assert len(seqlens) > 0, seqlens
        masks = [
            [x >= seqlen - self.sliding_window for x in range(seqlen)]
            for seqlen in seqlens
        ]
        to_cache_mask = torch.tensor(
            sum(masks, []), device=self.device, dtype=torch.bool
        )

        cached_elements = torch.tensor(
            [sum(mask) for mask in masks], device=self.device, dtype=torch.long
        )

        positions = torch.cat(
            [
                torch.arange(pos, pos + seqlen)
                for pos, seqlen in zip(seqpos, seqlens)
            ]
        ).to(device=self.device, dtype=torch.long)

        batch_idx = torch.tensor(
            sum([[i] * seqlen for i, seqlen in enumerate(seqlens)], []),
            device=self.device,
            dtype=torch.long,
        )

        cache_positions = (
            positions % self.sliding_window + batch_idx * self.sliding_window
        )

        first_prefill = seqpos[0] == 0
        subsequent_prefill = any(seqlen > 1 for seqlen in seqlens)

        if first_prefill:
            assert all([pos == 0 for pos in seqpos]), seqpos
            mask = BlockDiagonalCausalMask.from_seqlens(
                seqlens
            ).make_local_attention(self.sliding_window)

        elif subsequent_prefill:
            mask = BlockDiagonalMask.from_seqlens(
                q_seqlen=seqlens,
                kv_seqlen=[
                    s + cached_s.clamp(max=self.sliding_window).item()
                    for (s, cached_s) in zip(seqlens, self.kv_seqlens)
                ],
            ).make_local_attention_from_bottomright(self.sliding_window)
        else:
            mask = BlockDiagonalCausalWithOffsetPaddedKeysMask.from_seqlens(
                q_seqlen=seqlens,
                kv_padding=self.sliding_window,
                kv_seqlen=(self.kv_seqlens + cached_elements)
                .clamp(max=self.sliding_window)
                .tolist(),
            )

        return RotatingCacheInputMetadata(
            positions=positions,
            to_cache_mask=to_cache_mask,
            cached_elements=cached_elements,
            cache_positions=cache_positions[to_cache_mask],
            prefill=first_prefill or subsequent_prefill,
            mask=mask,
            seqlens=seqlens,
        )



# Copyright (c) 2022 Agora
# Licensed under The MIT License [see LICENSE for details]
from zeta.models.andromeda import Andromeda
from zeta.models.base import BaseModel
from zeta.models.gpt4 import GPT4, GPT4MultiModal
from zeta.models.llama import LLama2
from zeta.models.max_vit import MaxVit
from zeta.models.mega_vit import MegaVit
from zeta.models.palme import PalmE
from zeta.models.vit import ViT
from zeta.models.navit import NaViT


__all__ = [
    "BaseModel",
    "ViT",
    "MaxVit",
    "MegaVit",
    "PalmE",
    "GPT4",
    "GPT4MultiModal",
    "LLama2",
    "Andromeda",
    "NaViT",
]



import torch
from torch import nn
from einops import rearrange


def exists(val):
    return val is not None


class Residual(nn.Module):
    def __init__(self, dim, scale_residual=False, scale_residual_constant=1.0):
        super().__init__()
        self.residual_scale = (
            nn.Parameter(torch.ones(dim)) if scale_residual else None
        )
        self.scale_residual_constant = scale_residual_constant

    def forward(self, x, residual):
        if exists(self.residual_scale):
            residual = residual * self.residual_scale

        if self.scale_residual_constant != 1:
            residual = residual * self.scale_residual_constant

        return x + residual


class GRUGating(nn.Module):
    """
    GRUGating
    Overview: GRU gating mechanism

    Args:
        dim (int): dimension of the embedding
        scale_residual (bool): whether to scale residual
        kwargs (dict): keyword arguments for nn.GRUCell

    Atrributes:
        gru (nn.GRUCell): GRU cell
        residual_scale (nn.Parameter): residual scale

    Usage:
        >>> x = torch.randn(1, 10, 512)
        >>> gru_gating = GRUGating(dim=512)
        >>> gru_gating(x, x).shape
        torch.Size([1, 10, 512])

    """

    def __init__(self, dim, scale_residual=False, **kwargs):
        super().__init__()
        self.gru = nn.GRUCell(dim, dim)
        self.residual_scale = (
            nn.Parameter(torch.ones(dim)) if scale_residual else None
        )

    def forward(self, x, residual):
        """Forward method of GRUGating"""
        if exists(self.residual_scale):
            residual = residual * self.residual_scale

        gated_output = self.gru(
            rearrange(x, "b n d -> (b n) d"),
            rearrange(residual, "b n d -> (b n) d"),
        )

        return gated_output.reshape_as(x)



import torch
from torch import nn


class SimpleDecisionTree(nn.Module):
    """
    Simple decision tree model with residual connections and multi head output.


    Args:
        input_size (int): Input size of the model
        output_size (int): Output size of the model
        depth (int): Number of residual blocks
        heads (int): Number of output heads

    Example:
        >>> model = SimpleDecisionTree(
                input_size=10,
                output_size=5,
                depth=4,
                heads=3
            )
        >>> x = torch.randn(4, 10)
        >>> output = model(x)
        >>> print(output)
        [tensor([[-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436]],
                grad_fn=<AddmmBackward>), tensor([[-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436]],
                grad_fn=<AddmmBackward>), tensor([[-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436],
                 [-0.1015, -0.0114,  0.0370,  0.1362,  0.0436]],
                grad_fn=<AddmmBackward>)]
    """

    def __init__(
        self, input_size: int, output_size: int, depth: int, heads: int
    ):
        super(SimpleDecisionTree, self).__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.depth = depth
        self.heads = heads

        # Initial input layer
        self.input_layer = nn.Linear(input_size, input_size)

        # Residual blocks with batch norm and dropout
        self.residual_blocks = nn.ModuleList([])
        for _ in range(depth):
            layers = nn.Sequential(
                nn.Linear(input_size, input_size),
                nn.BatchNorm1d(input_size),
                nn.ReLU(),
                nn.Dropout(0.5),
                nn.Linear(input_size, input_size),
                nn.BatchNorm1d(input_size),
                nn.ReLU(),
            )
            self.residual_blocks.append(layers)

        # Recurrent layer for temproal dynamics
        self.recurrent_layer = nn.LSTM(input_size, input_size, batch_first=True)

        # Multi head output system
        self.output_heads = nn.ModuleList(
            [nn.Linear(input_size, output_size) for _ in range(heads)]
        )

    def forward(self, x: torch.Tensor):
        """Forward pass of the model.

        Args:
            x (torch.Tensor): _description_

        Returns:
            _type_: _description_
        """
        x = self.input_layer(x)

        # Applying residual connections
        for block in self.residual_blocks:
            residual = x
            x = block(x) + residual

        # Recurrent layer
        x, _ = self.recurrent_layer(x.unsqueeze(0))
        x = x.squeeze(0)

        # Multi head output
        outputs = [head(x) for head in self.output_heads]
        return outputs


# # Params
# input_size = 10
# output_size = 5
# depth = 4
# heads = 3
# batch_size = 4

# # model
# model = SimpleDecisionTree(
#     input_size,
#     output_size,
#     depth,
#     heads
# )

# x = torch.randn(batch_size, input_size)

# output = model(x)
# print(output)



from einops import rearrange
from torch import nn


class RNN(nn.Module):
    """
    Recurrent Neural Network for MNIST classification.

    Usage:
        net = RNN(
            ntoken=10,
            ninp=20,
            nhid=50,
            nlayers=2,
            dropout=0.5,
        )
        net(x)

    """

    def __init__(self, ntoken, ninp, nhid, nlayers, dropout=0.5):
        super(RNN, self).__init__()

        self.drop = nn.Dropout(p=dropout)
        self.encoder = nn.Embedding(ntoken, ninp)
        self.rnn = nn.LSTM(ninp, nhid, nlayers, dropout=dropout)
        self.decoder = nn.Linear(nhid, ntoken)

    def forward(self, input, hidden):
        """
        Take a forward pass through the network.

        Usage:
            net = RNN()
            net(x)
        """
        t, b = input.shape
        emb = self.drop(self.encoder(input))
        output, hidden = self.rnn(emb, hidden)
        output = rearrange(
            self.drop(output),
            "t b nhid -> (t b) nhid",
        )
        decoded = rearrange(
            self.decoder(output), "(t b) token -> t b token", t=t, b=b
        )
        return decoded, hidden



from zeta.structs.attn_layers import AttentionLayers


class Decoder(AttentionLayers):
    def __init__(self, **kwargs):
        assert "causal" not in kwargs, "cannot set causality on decoder"
        super().__init__(causal=True, **kwargs)



from einops import rearrange


def batched_dot_product(a, b):
    return rearrange(a * b, "b d -> b (d)").sum(dim=-1)


# x = torch.rand(1, 3)
# model = batched_dot_product(x, x)
# print(model.shape)



import math
from functools import partial
from itertools import zip_longest
from typing import Tuple

import torch
import torch.nn.functional as F
from einops import rearrange, repeat
from einops.layers.torch import Rearrange
from torch import nn
from vector_quantize_pytorch import RandomProjectionQuantizer

from zeta.structs.attn_layers import rotate_half
from zeta.nn.attention.attend import Attend
from zeta.nn.attention.local_attention_mha import LocalMHA
from zeta.nn.embeddings.rope import RotaryEmbedding

# constants
mlist = nn.ModuleList

Linear = partial(nn.Linear, bias=False)

LocalMHA = partial(LocalMHA, causal=True, prenorm=True)

# helper functions


def exists(val):
    return val is not None


def is_power_of_two(n):
    return math.log2(n).is_integer()


def all_unique(arr):
    return len(set(arr)) == len(arr)


def apply_fns(fns, tensors):
    return [fn(tensor) for fn, tensor in zip(fns, tensors)]


def cast_tuple(t, length=1):
    return t if isinstance(t, tuple) else ((t,) * length)


def default(*vals):
    for val in vals:
        if exists(val):
            return val
    return None


def eval_decorator(fn):
    def inner(model, *args, **kwargs):
        was_training = model.training
        model.eval()
        out = fn(model, *args, **kwargs)
        model.train(was_training)
        return out

    return inner


# sampling helpers


def log(t, eps=1e-20):
    return t.clamp(min=eps).log()


def gumbel_noise(t):
    noise = torch.zeros_like(t).uniform_(0, 1)
    return -log(-log(noise))


def gumbel_sample(t, temperature=1.0, dim=-1):
    return ((t / max(temperature, 1e-10)) + gumbel_noise(t)).argmax(dim=dim)


def top_k(logits, thres=0.9):
    k = int((1 - thres) * logits.shape[-1])
    val, ind = torch.topk(logits, k)
    probs = torch.full_like(logits, -torch.finfo(logits.dtype).max)
    probs.scatter_(1, ind, val)
    return probs


# rotary positional embedding w/ xpos
# https://arxiv.org/abs/2104.09864
# https://arxiv.org/abs/2212.10554v1


def apply_rotary_pos_emb(pos, t, scale=1.0):
    seq_len = t.shape[-2]

    pos = pos[..., -seq_len:, :]
    if not isinstance(scale, (int, float)):
        scale = scale[..., -seq_len:, :]

    return (t * pos.cos() * scale) + (rotate_half(t) * pos.sin() * scale)


def apply_rotary_pos_emb_qk(rotary_emb, q, k):
    freqs, scale = rotary_emb
    q = apply_rotary_pos_emb(freqs, q, scale)
    k = apply_rotary_pos_emb(freqs, k, scale**-1)
    return q, k


# token shift, from Peng et al of RWKV


def token_shift(t):
    t, t_shift = t.chunk(2, dim=-1)
    t_shift = F.pad(t_shift, (0, 0, 1, -1))
    return torch.cat((t, t_shift), dim=-1)


# hierarchy related classes


def pad_seq_to_multiple(t, mult):
    seq_len = t.shape[-2]
    next_seq_len_mult = math.ceil(seq_len / mult) * mult
    remainder = next_seq_len_mult - seq_len

    if remainder == 0:
        return t, seq_len

    t = F.pad(t, (0, 0, 0, remainder), value=0.0)
    return t, seq_len


def curtail_seq_to_multiple(t, mult):
    seq_len = t.shape[-2]
    prev_seq_len_mult = (seq_len // mult) * mult
    remainder = seq_len - prev_seq_len_mult

    if remainder == 0:
        return t

    t = t[..., :prev_seq_len_mult, :]
    return t


def hierarchical_cat(tokens, strides: Tuple[int, ...]):
    assert len(tokens) == len(strides)

    if all([s == 1 for s in strides]):
        return torch.cat(tokens, dim=-1)

    tokens = [
        repeat(t, "b n d -> b (n s) d", s=s) for t, s in zip(tokens, strides)
    ]
    min_seq_len = min([t.shape[-2] for t in tokens])
    tokens = [t[..., :min_seq_len, :] for t in tokens]
    return torch.cat(tokens, dim=-1)


class CausalConv(nn.Module):
    def __init__(self, dim_in, dim_out, kernel_size, stride=1):
        super().__init__()
        self.causal_padding = kernel_size - 1
        self.conv = nn.Conv1d(dim_in, dim_out, kernel_size, stride=stride)

    def forward(self, x):
        x = F.pad(x, (self.causal_padding, 0))
        return self.conv(x)


class Compress(nn.Module):
    def __init__(
        self,
        *,
        dim,
        dim_out,
        num_tokens=None,
        stride=1,
        compress_factor=1,
        expansion_factor=4,
        dim_head=64,
        heads=8,
        ignore_index=0,
        should_recon=False,
        should_prophet=False,
        prophet_num_predictions=None,
    ):
        super().__init__()
        assert compress_factor > 0 and is_power_of_two(compress_factor)

        self.stride = stride
        self.no_compress = compress_factor == 1
        self.compress_factor = compress_factor

        self.should_recon = should_recon
        self.should_prophet = should_prophet

        if self.no_compress:
            self.compress_fn = (
                Linear(dim, dim_out) if dim != dim_out else nn.Identity()
            )
            return

        dim_inner = int(dim * expansion_factor)

        self.compress_fn = nn.Sequential(
            Rearrange("b n d -> b d n"),
            CausalConv(dim, dim_inner, compress_factor, stride=stride),
            nn.SiLU(),
            nn.Conv1d(dim_inner, dim_out, 1),
            Rearrange("b d n -> b n d"),
        )

        if should_recon:
            assert exists(num_tokens)
            self.to_recon = Linear(dim_out, compress_factor * num_tokens)

        if should_prophet:
            assert exists(prophet_num_predictions)
            self.to_prophet = Linear(dim_out, prophet_num_predictions)

        self.ignore_index = ignore_index

    def prophet(self, h, ids):
        if not self.should_prophet:
            return torch.zeros((), device=h.device).requires_grad_()

        c = self.compress_factor
        seq_len = ids.shape[-1]

        prophet_logits = self.to_prophet(h)
        prophet_logits = rearrange(
            prophet_logits, "b n (c d) -> (b c) d n", c=c
        )

        prophet_ids = F.pad(ids, (-1, c), value=self.ignore_index)
        prophet_ids = tuple(prophet_ids[:, i : (seq_len + i)] for i in range(c))
        prophet_ids = torch.stack(prophet_ids, dim=1)
        prophet_ids = rearrange(prophet_ids, "b c n -> (b c) n")

        if self.stride > 1:
            prophet_ids = prophet_ids[..., :: self.stride]

        prophet_loss = F.cross_entropy(
            prophet_logits, prophet_ids, ignore_index=self.ignore_index
        )
        return prophet_loss

    def recon(self, h, ids):
        assert self.should_recon

        if self.no_compress:
            return torch.zeros((), device=h.device).requires_grad_()

        c = self.compress_factor
        seq_len = ids.shape[-1]

        recon_logits = self.to_recon(h)
        recon_logits = rearrange(recon_logits, "b n (c d) -> (b c) d n", c=c)

        recon_ids = F.pad(ids, (c - 1, 0), value=self.ignore_index)
        recon_ids = tuple(recon_ids[:, i : (seq_len + i)] for i in range(c))
        recon_ids = torch.stack(recon_ids, dim=1)
        recon_ids = rearrange(recon_ids, "b c n -> (b c) n")

        if self.stride > 1:
            recon_ids = recon_ids[..., :: self.stride]

        recon_loss = F.cross_entropy(
            recon_logits, recon_ids, ignore_index=self.ignore_index
        )
        return recon_loss

    def forward(self, x):
        return self.compress_fn(x)


class HierarchicalMerge(nn.Module):
    def __init__(self, dims: Tuple[int, ...], dim_out, h_strides=1):
        super().__init__()
        dim = sum(dims)

        strides = cast_tuple(h_strides, len(dims))
        assert len(strides) == len(dims)

        self.strides = strides

        self.net = nn.Sequential(
            RMSNorm(dim),
            nn.Linear(dim, dim_out * 2),
            nn.SiLU(),
            nn.Linear(dim_out * 2, dim_out),
        )

    def forward(self, tokens):
        x = hierarchical_cat(tokens, self.strides)
        return self.net(x)


# classes


class RMSNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.scale = dim**0.5
        self.gamma = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        return F.normalize(x, dim=-1) * self.scale * self.gamma


class FeedForward(nn.Module):
    def __init__(self, dim, mult=4):
        super().__init__()
        dim_inner = int(dim * mult)

        self.net = nn.Sequential(
            RMSNorm(dim),
            Linear(dim, dim_inner),
            nn.GELU(),
            Linear(dim_inner, dim),
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, dim, dim_head=64, heads=8, use_flash_attn=False):
        super().__init__()
        self.scale = dim_head**-0.5
        self.heads = heads
        dim_inner = dim_head * heads

        self.norm = RMSNorm(dim)
        self.rotary_emb = RotaryEmbedding(dim_head)

        self.attend = Attend(causal=True, use_flash_attn=use_flash_attn)

        self.to_qkv = Linear(dim, dim_inner * 3)
        self.to_out = Linear(dim_inner, dim)

    def forward(self, x):
        n = x.shape[-2]
        x = self.norm(x)

        q, k, v = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.heads),
            (q, k, v),
        )

        rotary_emb = self.rotary_emb(n)
        q, k = apply_rotary_pos_emb_qk(rotary_emb, q, k)

        out = self.attend(q, k, v)

        out = rearrange(out, "b h n d -> b n (h d)")
        return self.to_out(out)


class HierarchicalBlock(nn.Module):
    def __init__(
        self,
        dim,
        dim_head=64,
        heads=8,
        window_size=None,
        compress_factor=1,
        stride=1,
        ff_mult=4,
    ):
        super().__init__()
        self.stride = stride

        assert is_power_of_two(compress_factor)
        self.compress_factor = compress_factor
        self.no_compress = compress_factor == 1

        assert not exists(window_size) or window_size >= 0
        self.has_attn = window_size != 0

        self.attn = None

        if self.has_attn:
            attn_klass = Attention
            if exists(window_size):
                attn_klass = partial(LocalMHA, window_size=window_size)

            self.attn = attn_klass(dim=dim, dim_head=dim_head, heads=heads)

        self.ff = FeedForward(dim=dim, mult=ff_mult)

    def forward(self, x):
        c = self.compress_factor
        axial_dim = c // self.stride

        x, orig_seq_len = pad_seq_to_multiple(x, axial_dim)

        # hierarchical attention is performed with a simple axial attention

        # this, and using a convolution for compressing at the beginning
        # is one of the improvements on top of hourglass transformer
        # the downside is that the savings are only O(c) instead of O(c ** 2) as in hourglass transformer
        # you can get the O(c ** 2) saving by setting the hierarchical stride
        # == c, but you'll see that performance is much worse, as some tokens
        # will have a c - 1 token gap to the last hierarchical token

        if not self.no_compress:
            x = rearrange(x, "b (n c) d -> (b c) n d", c=axial_dim)

        if exists(self.attn):
            x = self.attn(token_shift(x)) + x

        x = self.ff(token_shift(x)) + x

        if not self.no_compress:
            x = rearrange(x, "(b c) n d -> b (n c) d", c=axial_dim)

        return x[:, :orig_seq_len]


class HierarchicalTransformer(nn.Module):
    def __init__(
        self,
        *,
        num_tokens,
        dim,
        depth,
        seq_len=2048,
        dim_head=64,
        heads=8,
        ff_mult=4,
        hierarchies=1,
        window_sizes=None,
        hierarchical_stride=1,
        hierarchy_merge_all=False,
        # whether to pass the pooled hierarchical information back to all
        # hierarchies or just one doing the prediction
        ignore_index=0,
        use_flash_attn=False,
        recon_loss_weight=0.1,
        prophet_loss_weight=0.0,
        prophet_loss_use_quantized=False,
        # for prophet, whether to use the next 1x token ids, or use the ids
        # from random projection quantization
        prophet_quantized_use_embed=False,
        predict_hierarchy=None,
        predict_use_all_hierarchy=False,
        rq_num_codebooks=4,
        rq_codebook_dim=256,
        rq_codebook_size=1024,
    ):
        """
        By not specifying hierarchies and window_sizes, you basically default to a regular autoregressive transformer with attention across full sequence length,
        Three hierarchies, all servicing predicting the next token

        from zeta.nn import HierarchicalTransformer

        model = HierarchicalTransformer(
            num_tokens = 256,
            dim = (128, 256, 512, 1024),
            depth = 8,
            seq_len = 1024,
            use_flash_attn = True,
            ff_mult = (2, 2, 4, 4),
            dim_head = (16, 32, 64, 64),
            heads = (2, 4, 8, 8),
            hierarchies = (1, 2, 4, 16),
            hierarchical_stride = (1, 1, 1, 8),  # this would determine the stride when compressing, and when concatting the hierarchical tokens to the fine tokens, the past tokens will be repeated this amount of time. causality is not violated as using the trick from hourglass transformers where sequence is shifted by compression factor - 1. recommend sticking with 1 except for highly compressed hierarchies, as it becomes very uncompetitive with baseline and generations look off
            window_sizes = (16, 32, 64, None)
        ).cuda()

        # hierarchies
        # 1x - dim 128 - attention (2 heads, 16 dim, receptive field 16)
        # 2x - dim 256 - attention (4 heads, 32 dim, receptive field 32)
        # 4x - dim 512 - attention (8 heads, 64 dim, receptive field 64)
        # 8x - dim 1024 - attention (8 heads, 64 dim, receptive field of all)
        """
        super().__init__()
        self.seq_len = seq_len

        hierarchies = cast_tuple(hierarchies)
        assert all_unique(
            hierarchies
        ), "hierarchies compression factors must be all unique integers"
        assert all(
            [*map(is_power_of_two, hierarchies)]
        ), "only powers of two allowed for hierarchies"

        self.hierarchies = hierarchies

        # just use a simple tuple list per hyperparameter to customize each
        # hierarchy

        num_hierarchies = len(hierarchies)

        dims = cast_tuple(dim, num_hierarchies)
        assert len(dims) == num_hierarchies

        window_sizes = cast_tuple(window_sizes, num_hierarchies)
        assert len(window_sizes) == num_hierarchies

        dim_head = cast_tuple(dim_head, num_hierarchies)
        assert len(dim_head) == num_hierarchies

        heads = cast_tuple(heads, num_hierarchies)
        assert len(heads) == num_hierarchies

        ff_mult = cast_tuple(ff_mult, num_hierarchies)
        assert len(ff_mult) == num_hierarchies

        hierarchical_stride = cast_tuple(hierarchical_stride, num_hierarchies)

        assert all(
            [*map(is_power_of_two, hierarchical_stride)]
        ), "all hierarchical strides must be power of two"
        assert all(
            [s <= h for s, h in zip(hierarchical_stride, hierarchies)]
        ), (
            "all strides must be less than the compression factor of the"
            " hierarchy"
        )

        self.h_strides = hierarchical_stride

        assert len(hierarchical_stride) == num_hierarchies

        # this determines to which hierarchy is everything pooled into for final prediction
        # however, final next token prediction can also use all hierarchies
        # with `predict_use_all_hierarchy`

        predict_hierarchy = default(predict_hierarchy, min(hierarchies))
        self.predict_hierarchy_index = hierarchies.index(predict_hierarchy)
        hierarchy_predict_dim = dims[self.predict_hierarchy_index]

        self.hierarchy_merge_all = hierarchy_merge_all
        assert (
            hierarchy_merge_all
            or self.h_strides[self.predict_hierarchy_index] == 1
        ), (
            "the hierarchy level being used for final next token prediction"
            " must have compression stride of 1"
        )

        # training related loss weights

        self.recon_loss_weight = recon_loss_weight
        self.prophet_loss_weight = prophet_loss_weight

        should_recon = recon_loss_weight > 0
        should_prophet = prophet_loss_weight > 0

        self.should_recon = should_recon
        self.should_prophet = should_prophet

        self.prophet_loss_use_quantized = prophet_loss_use_quantized
        self.prophet_quantized_use_embed = prophet_quantized_use_embed

        # token embedding

        dim_token_emb = max(dims)
        self.token_emb = nn.Embedding(num_tokens, dim_token_emb)

        # hierarchy compressions - 1x just uses the base token_emb weights

        self.compressors = mlist([])

        for dim, hierarchy, stride in zip(
            dims, hierarchies, hierarchical_stride
        ):
            self.compressors.append(
                Compress(
                    dim=dim_token_emb,
                    dim_out=dim,
                    num_tokens=num_tokens,
                    compress_factor=hierarchy,
                    stride=stride,
                    should_recon=should_recon,
                    should_prophet=should_prophet,
                    prophet_num_predictions=(
                        (hierarchy * num_tokens)
                        if not prophet_loss_use_quantized
                        else (rq_num_codebooks * rq_codebook_size)
                    ),
                )
            )

        # post token embedding norms

        self.post_token_emb_norms = mlist([nn.LayerNorm(dim) for dim in dims])

        # layers

        self.layers = mlist([])

        self.dims = dims

        self.hierarchical_merges = mlist([])
        self.need_hierarchical_merge = num_hierarchies > 1

        for _ in range(depth):
            hierarchical_layer = mlist([])

            # add a transformer block for each layer in the hierarchy

            for (
                hierarchy,
                h_stride,
                h_dim,
                h_window_size,
                h_dim_head,
                h_heads,
                h_ff_mult,
            ) in zip(
                hierarchies,
                hierarchical_stride,
                dims,
                window_sizes,
                dim_head,
                heads,
                ff_mult,
            ):
                # make sure the window size never exceeds the effective
                # sequence length

                effective_seq_len = seq_len // hierarchy

                if exists(h_window_size) and h_window_size > effective_seq_len:
                    print(
                        f"window size for hierarchy {hierarchy}x is greater"
                        " than effective sequence length - setting window size"
                        " to None (which would use normal full attention)"
                    )
                    h_window_size = None

                # add attention and feedforward

                hierarchical_layer.append(
                    HierarchicalBlock(
                        dim=h_dim,
                        dim_head=h_dim_head,
                        heads=h_heads,
                        window_size=h_window_size,
                        compress_factor=hierarchy,
                        stride=h_stride,
                        ff_mult=h_ff_mult,
                    )
                )

            self.layers.append(hierarchical_layer)

            # for merging the information across hierarchies
            # for now, only one direction, from all hierarchies to the
            # hierarchy that is being used to make predictions on, set by
            # predict_hierarchy_index above

            if not self.need_hierarchical_merge:
                continue

            merge = HierarchicalMerge(
                dims=dims,
                dim_out=(
                    hierarchy_predict_dim
                    if not self.hierarchy_merge_all
                    else sum(dims)
                ),
                h_strides=hierarchical_stride,
            )

            self.hierarchical_merges.append(merge)

        # final post-transformer norms, for all hierarchies

        self.norms = mlist([nn.LayerNorm(dim) for dim in dims])

        # random projection quantizer, for another approach to hierarchical
        # predictive coding

        if self.prophet_loss_use_quantized:
            rpq_klass = partial(
                RandomProjectionQuantizer,
                num_codebooks=rq_num_codebooks,
                codebook_dim=rq_codebook_dim,
                codebook_size=rq_codebook_size,
            )

            self.rand_proj_quantizers = mlist(
                [rpq_klass(dim=dim) for dim in dims]
            )
            self.rq_num_codebooks = rq_num_codebooks

        # to logit, for hierarchy set at predict_hierarchy_index, or all
        # hierarchies

        self.predict_use_all_hierarchy = predict_use_all_hierarchy
        logit_dim_in = (
            sum(dims) if predict_use_all_hierarchy else hierarchy_predict_dim
        )

        self.to_logits = Linear(logit_dim_in, num_tokens)

        # training related loss parameters

        self.ignore_index = ignore_index

    @torch.no_grad()
    @eval_decorator
    def generate(
        self, prompt, seq_len, temperature=1.0, filter_thres=0.9, **kwargs
    ):
        b, t, device = *prompt.shape, prompt.device

        out = prompt

        for _ in range(seq_len):
            logits = self.forward(out[:, -self.seq_len :], **kwargs)[:, -1]
            filtered_logits = top_k(logits, thres=filter_thres)
            sample = gumbel_sample(filtered_logits, temperature=temperature)
            sample = rearrange(sample, "b -> b 1")
            out = torch.cat((out, sample), dim=-1)

        return out[:, t:]

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(
        self,
        ids,
        return_loss=False,
        return_hierarchical_token_embeds=False,
        return_hierarchical_embeds=False,
        ablate_hierarchical_merge=False,
        return_random_proj_quantize_ids=False,
    ):
        """
        einops notation:

        b - batch
        n - sequence length
        c - compression factor
        d - dimension
        """

        # if training, predict next token in sequence

        if return_loss:
            ids, labels = ids[:, :-1], ids[:, 1:]

        # assert seq len

        assert ids.shape[-1] <= self.seq_len

        # get token embeddings, and pad to multiple of compression factor

        x = self.token_emb(ids)

        # for every hierarchy, compress token embeddings appropriately to the
        # hierarchical embeddings

        tokens = []

        for compress in self.compressors:
            tokens.append(compress(x))

        # save hierarchical tokens right before norm for random projection
        # quantization, if needed

        post_compressed_tokens = tokens

        # post embedding norms

        tokens = apply_fns(self.post_token_emb_norms, tokens)

        # if one wants all the compressed token embeds
        # just to investigate the space

        if return_hierarchical_token_embeds:
            return tokens

        # layers

        for layer, merge in zip_longest(self.layers, self.hierarchical_merges):
            tokens = apply_fns(layer, tokens)

            # pool the information all hierarchies
            # and then update the tokens that will be used to make the final
            # autoregressive prediction

            if not self.need_hierarchical_merge or ablate_hierarchical_merge:
                continue

            pooled = merge(tokens)

            if self.hierarchy_merge_all:
                tokens = [
                    (t + p[..., ::s, :])
                    for t, p, s in zip(
                        tokens, pooled.split(self.dims, dim=-1), self.h_strides
                    )
                ]
            else:
                predict_tokens = tokens[self.predict_hierarchy_index]
                predict_tokens = predict_tokens + pooled
                tokens[self.predict_hierarchy_index] = predict_tokens

        # final normalized embeddings

        embeds = apply_fns(self.norms, tokens)

        # if the researcher wants the randomly projected ids of either
        # compressed tokens or embeddings of the hierarchies

        if return_random_proj_quantize_ids:
            assert self.prophet_loss_use_quantized

            quantize_input = (
                embeds
                if self.prophet_quantized_use_embed
                else post_compressed_tokens
            )
            hierarchical_ids = apply_fns(
                self.rand_proj_quantizers, quantize_input
            )
            return hierarchical_ids

        # if one wants all the normalized hierarchical embeds

        if return_hierarchical_embeds:
            return embeds

        # select the hierarchical embeddings that will be doing the predicting

        if self.predict_use_all_hierarchy:
            predict_embed = hierarchical_cat(embeds, self.h_strides)
        else:
            predict_embed = embeds[self.predict_hierarchy_index]

        # logits for predicting next token

        logits = self.to_logits(predict_embed)

        if not return_loss:
            return logits

        ce_loss_fn = partial(F.cross_entropy, ignore_index=self.ignore_index)

        # autoregressive loss (predictive coding)

        logits = rearrange(logits, "b n c -> b c n")
        ce_loss = ce_loss_fn(logits, labels)

        # reconstruction losses for hierarchy tokens

        recon_losses = prophet_losses = torch.zeros(
            (), device=self.device
        ).requires_grad_()

        if self.should_recon:
            for compress, t in zip(self.compressors, embeds):
                recon_loss = compress.recon(t, ids)
                recon_losses = recon_losses + recon_loss

        # prophet losses for hierarchy tokens

        if self.should_prophet:
            if self.prophet_loss_use_quantized:
                # using random projected quantizer of the next hierarchical
                # token

                quantize_input = (
                    embeds
                    if self.prophet_quantized_use_embed
                    else post_compressed_tokens
                )

                hierarchical_ids = apply_fns(
                    self.rand_proj_quantizers, quantize_input
                )

                for hierarchy, stride, compress, embed, pred_ids in zip(
                    self.hierarchies,
                    self.h_strides,
                    self.compressors,
                    embeds,
                    hierarchical_ids,
                ):
                    if hierarchy == 1:
                        continue

                    prophet_logits = compress.to_prophet(embed)

                    axial_dim = hierarchy // stride

                    prophet_logits = curtail_seq_to_multiple(
                        prophet_logits, axial_dim
                    )
                    pred_ids = curtail_seq_to_multiple(pred_ids, axial_dim)

                    prophet_logits, pred_ids = map(
                        lambda t: rearrange(
                            t, "b (n c) ... -> (b c) n ...", c=axial_dim
                        ),
                        (prophet_logits, pred_ids),
                    )

                    prophet_logits = rearrange(
                        prophet_logits[:, :-1],
                        "b n (q c) -> (b q) c n",
                        q=self.rq_num_codebooks,
                    )
                    pred_ids = rearrange(pred_ids[:, 1:], "b n q -> (b q) n")

                    prophet_loss = ce_loss_fn(prophet_logits, pred_ids)
                    prophet_losses = prophet_losses + prophet_loss

            else:
                # or predicting the next N 1x base token ids
                # like prophetnet paper

                for compress, t in zip(self.compressors, embeds):
                    prophet_loss = compress.prophet(t, ids)
                    prophet_losses = prophet_losses + prophet_loss

        # total loss

        total_loss = (
            ce_loss
            + recon_losses * self.recon_loss_weight
            + prophet_losses * self.prophet_loss_weight
        )

        return total_loss, (ce_loss, recon_losses, prophet_losses)



import torch
import torch.nn as nn

device = "cuda:0"
dtype = torch.float16


class ParallelWrapper:
    """
    A simple wrapper to enable easy usage of data parallelism.

    Arguments:
        model: The neural network model to be parallelized.
        device (optional): The device to which the model should be moved. Default: "cuda".
        use_data_parallel (optional): A boolean flag to indicate whether to use data parallelism or not. Default: True.

    Usage:
        The `ParallelWrapper` class can be used as a wrapper for neural networks and is especially suited for transformer architectures.

        Example:
            model = nn.Linear(512, 512)
            model = ParallelWrapper(model)

        This will return the model wrapped in the `ParallelWrapper` class. The `use_data_parallel` parameter allows for switching on data parallelism.
    """

    def __init__(self, model, device="cuda", use_data_parallel=True):
        self.model = model.to(device)
        self.use_data_parallel = use_data_parallel
        self.device = device

        if self.use_data_parallel and torch.cuda.device_count() < 1:
            print(f"Using {torch.cuda.device_count()} GPUS")
            self.model = nn.DataParallel(self.model)

    def forward(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def to(self, device):
        self.device = device
        self.model = self.model.to(device)
        return self

    def __getattr__(self, name):
        # redirect attribute access to the internal model to allow direct access to its methods and props
        return getattr(self.model, name)



import argparse
from zeta.cloud.main import zetacloud


def main():
    """Main function for the CLI

    Args:
        task_name (str, optional): _description_. Defaults to None.
        cluster_name (str, optional): _description_. Defaults to "[ZetaTrainingRun]".
        cloud (Any, optional): _description_. Defaults to AWS().
        gpus (str, optional): _description_. Defaults to None.

    Examples:
        $ zetacloud -t "test" -c "[ZetaTrainingRun]" -cl AWS -g "1 V100"


    """
    parser = argparse.ArgumentParser(description="Zetacloud CLI")
    parser.add_argument("-t", "--task_name", type=str, help="Task name")
    parser.add_argument(
        "-c",
        "--cluster_name",
        type=str,
        default="[ZetaTrainingRun]",
        help="Cluster name",
    )
    parser.add_argument(
        "-cl", "--cloud", type=str, default="AWS", help="Cloud provider"
    )
    parser.add_argument("-g", "--gpus", type=str, help="GPUs")
    parser.add_argument(
        "-f", "--filename", type=str, default="train.py", help="Filename"
    )
    parser.add_argument("-s", "--stop", action="store_true", help="Stop flag")
    parser.add_argument("-d", "--down", action="store_true", help="Down flag")
    parser.add_argument(
        "-sr", "--status_report", action="store_true", help="Status report flag"
    )

    # Generate API key
    # parser.add_argument(
    #     "-k", "--generate_api_key", action="store_true", help="Generate key flag"
    # )

    # Sign In
    # parser.add_argument(
    #     "-si", "--sign_in", action="store_true", help="Sign in flag"
    # )

    args = parser.parse_args()

    zetacloud(
        task_name=args.task_name,
        cluster_name=args.cluster_name,
        cloud=args.cloud,
        gpus=args.gpus,
        filename=args.filename,
        stop=args.stop,
        down=args.down,
        status_report=args.status_report,
    )


# if __name__ == "__main__":
#     main()



from typing import Tuple

import torch
from torch import nn

from zeta.nn.attention.attend import Attend
from zeta.nn.modules.cache import CacheView


def repeat_kv(keys: torch.Tensor, values: torch.Tensor, repeats: int, dim: int):
    keys = torch.repeat_interleave(keys, repeats=repeats, dim=dim)
    values = torch.repeat_interleave(values, repeats=repeats, dim=dim)
    return keys, values


def precompute_freqs_cis(
    dim: int, end: int, theta: float = 10000.0
) -> torch.Tensor:
    freqs = 1.0 / (
        theta ** (torch.arange(0, dim, 2)[: (dim // 2)].float() / dim)
    )
    t = torch.arange(end, device=freqs.device)  # type: ignore
    freqs = torch.outer(t, freqs).float()  # type: ignore
    return torch.polar(torch.ones_like(freqs), freqs)  # complex64


def apply_rotary_emb(
    xq: torch.Tensor,
    xk: torch.Tensor,
    freqs_cis: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    xq_ = torch.view_as_complex(xq.float().reshape(*xq.shape[:-1], -1, 2))
    xk_ = torch.view_as_complex(xk.float().reshape(*xk.shape[:-1], -1, 2))
    freqs_cis = freqs_cis[:, None, :]
    xq_out = torch.view_as_real(xq_ * freqs_cis).flatten(2)
    xk_out = torch.view_as_real(xk_ * freqs_cis).flatten(2)
    return xq_out.type_as(xq), xk_out.type_as(xk)


# mgqa
class MGQA(nn.Module):
    """
    Multi-Headed Generalized Query Attention

    Args:
        dim (int): Input dimension
        n_layers (int): Number of layers
        head_dim (int): Head dimension
        hidden_dim (int): Hidden dimension
        n_heads (int): Number of heads
        n_kv_heads (int): Number of key/value heads
        sliding_window (int): Sliding window size
        norm_eps (float): Epsilon for layer norm
        vocab_size (int): Vocabulary size
        attn_dropout (float): Dropout probability
        max_batch_size (int): Maximum batch size
        flash (bool): Use FlashAttention

    Usage:
    >>> model = MGQA(768, 12, 64, 2048, 8, 8, 512, 1e-6, 32000, 0.1, 0, False)
    >>> x = torch.randn(1, 768)
    >>> model(x).shape


    """

    def __init__(
        self,
        dim: int,
        n_layers: int,
        head_dim: int,
        hidden_dim: int,
        n_heads: int,
        n_kv_heads: int,
        sliding_window: int,
        norm_eps: float,
        vocab_size: int,
        attn_dropout: float = 0.0,  # moved to the end
        max_batch_size: int = 0,  # default argument
        flash: bool = False,  # non-default argument
    ):
        super().__init__()

        self.dim = dim
        self.n_layers = n_layers
        self.head_dim = head_dim
        self.hidden_dim = hidden_dim
        self.n_heads = n_heads
        self.n_kv_heads = n_kv_heads
        self.sliding_window = sliding_window
        self.norm_eps = norm_eps
        self.vocab_size = vocab_size
        self.max_batch_size = max_batch_size
        self.attn_dropout = attn_dropout
        self.flash = flash

        self.repeats = self.n_heads // self.n_kv_heads
        self.scale = self.head_dim**-0.5

        self.wq = nn.Linear(self.dim, self.n_heads * self.head_dim, bias=False)
        self.wk = nn.Linear(
            self.dim, self.n_kv_heads * self.head_dim, bias=False
        )
        self.wv = nn.Linear(
            self.n_heads * self.head_dim,
            self.n_kv_heads * self.head_dim,
            bias=False,
        )
        self.wo = nn.Linear(self.n_heads * self.head_dim, self.dim, bias=False)

        self.attn = Attend(
            dropout=self.attn_dropout,
            causal=True,
            flash=self.flash,
        )

    def forward(
        self,
        x: torch.Tensor,
        freqs_cis: torch.Tensor,
        cache: CacheView,
    ) -> torch.Tensor:
        """
        Forward pass

        Args:
            x (torch.Tensor): Input tensor
            freqs_cis (torch.Tensor): Precomputed frequencies
            cache (CacheView): Cache view

        Example:
        >>> model = MGQA(768, 12, 64, 2048, 8, 8, 512, 1e-6, 32000, 0.1, 0, False)
        >>> x = torch.randn(1, 768)
        >>> freqs_cis = torch.randn(1, 768)
        >>> cache = CacheView(1, 512, 8, 8, 64)
        >>> model(x, freqs_cis, cache).shape


        """
        seqlen_sum, _ = x.shape

        xq, xk, xv = self.wq(x), self.wk(x), self.wv(x)

        xq = xq.view(seqlen_sum, self.n_heads, self.head_dim)

        xk = xk.view(seqlen_sum, self.n_kv_heads, self.head_dim)

        xv = xv.view(seqlen_sum, self.n_kv_heads, self.head_dim)

        xq, xk = apply_rotary_emb(
            xq,
            xk,
            freqs_cis=freqs_cis,
        )

        if cache.prefill:
            key, val = cache.interleave_kv(xk, xv)
        else:
            cache.update(xk, xv)
            key, val = cache.keys, cache.values

            key = key.view(
                seqlen_sum * cache.sliding_window,
                self.n_kv_heads,
                self.head_dim,
            )

            val = val.view(
                seqlen_sum * cache.sliding_window,
                self.n_kv_heads,
                self.head_dim,
            )

        # repeat keys and values to match number of query heads
        key, val = repeat_kv(key, val, self.repeats, dim=1)

        # attention
        xq, key, val = xq[None, ...], key[None, ...], val[None, ...]
        output = self.attn(xq, key, val, self.scale)

        return self.wo(output.view_as(x))



import torch
from torch import nn

from zeta.utils.main import exists, l2norm


class AbsolutePositionalEmbedding(nn.Module):
    def __init__(self, dim, max_seq_len, l2norm_embed=False):
        super().__init__()
        self.scale = dim**-0.5 if not l2norm_embed else 1.0
        self.max_seq_len = max_seq_len
        self.l2norm_embed = l2norm_embed
        self.emb = nn.Embedding(max_seq_len, dim)

    def forward(self, x, pos=None):
        seq_len, device = x.shape[-1], x.device
        assert seq_len <= self.max_seq_len, (
            f"You are passing in a sequence length of {seq_len} but you"
            " absolute positional embedding has a max of length of"
            f" {self.max_seq_len}"
        )

        if not exists(pos):
            pos = torch.arange(seq_len, device=device)

        pos_emb = self.emb(pos)
        pos_emb = pos_emb * self.scale
        return l2norm(pos_emb) if self.l2norm_embed else pos_emb



import torch.nn as nn


class ImagePatchCreatorProjector(nn.Module):
    """
    Image Patch Creator and Projector Layer.

    This layer dynamically creates and projects image patches suitable for
    feeding into a transformer decoder. It is designed to handle input tensors
    of arbitrary shape and outputs a tensor of shape (B, SEQLEN, Dimension).

    Attributes:
        max_patch_size (int): The maximum size of each image patch.
        embedding_dim (int): The dimension of the output embeddings.
    """

    def __init__(self, max_patch_size, embedding_dim):
        """
        Initializes the ImagePatchCreatorProjector.

        Args:
            max_patch_size (int): The maximum size of each image patch.
            embedding_dim (int): The dimension of the output embeddings.
        """
        super().__init__()
        self.max_patch_size = max_patch_size
        self.embedding_dim = embedding_dim
        self.adaptive_pool = nn.AdaptiveAvgPool2d(
            (max_patch_size, max_patch_size)
        )
        self.projection = None

    def forward(self, x):
        """
        Forward pass of the layer.

        Args:
            x (torch.Tensor): The input tensor with shape (B, C, H, W).

        Returns:
            torch.Tensor: The output tensor with shape (B, SEQLEN, Dimension).
        """
        try:
            B, C, H, W = x.shape
            dynamic_patch_size = self.calculate_dynamic_patch_size(H, W)
            self.projection = nn.Linear(
                dynamic_patch_size * dynamic_patch_size * C, self.embedding_dim
            )

            x = self.create_patches(x, dynamic_patch_size)
            x = self.adaptive_pool(x)
            x = x.view(B, -1, dynamic_patch_size * dynamic_patch_size * C)
            x = self.projection(x)

            return x
        except Exception as e:
            # Handle exceptions and potentially log them
            print(f"Error during forward pass: {e}")
            return None

    def calculate_dynamic_patch_size(self, H, W):
        """
        Calculate dynamic patch size based on the dimensions of the input image.

        Args:
            H (int): Height of the input image.
            W (int): Width of the input image.

        Returns:
            int: Calculated patch size.
        """
        # Example logic; this can be adjusted based on specific requirements
        return min(H, W, self.max_patch_size)

    def create_patches(self, x, patch_size):
        """
        Create image patches from the input tensor.

        Args:
            x (torch.Tensor): The input tensor.
            patch_size (int): Size of each patch.

        Returns:
            torch.Tensor: Tensor with created patches.
        """
        B, C, H, W = x.shape
        x = x.unfold(2, patch_size, patch_size).unfold(
            3, patch_size, patch_size
        )
        x = x.contiguous().view(B, -1, patch_size, patch_size, C)
        x = (
            x.permute(0, 1, 4, 2, 3)
            .contiguous()
            .view(B, -1, patch_size, patch_size)
        )
        return x


# # Example Usage
# # Initialize the layer
# patch_projector = ImagePatchCreatorProjector(max_patch_size=16, embedding_dim=768)

# # Example input tensor (randomly generated for demonstration)
# input_tensor = torch.randn(1, 3, 64, 64)  # Shape: [B, C, H, W]

# # Forward pass
# output_tensor = patch_projector(input_tensor)
# print(
#     f"Output Shape: {output_tensor.shape if output_tensor is not None else 'Error in processing'}"
# )



import torch
from torch import nn


class VisionEmbedding(nn.Module):
    """
    Image to Patch Embedding


    Args:
        img_size (int): The image size.
        patch_size (int): The patch size.
        in_chans (int): The number of input channels.
        embed_dim (int): The embedding dimension.
        contain_mask_token (bool): Whether to contain mask token or not.
        prepend_cls_token (bool): Whether to prepend cls token or not.

    Attributes:
        patch_shape (tuple): The patch shape.
        img_size (tuple): The image size.
        patch_size (tuple): The patch size.
        num_patches (int): The number of patches.
        proj (nn.Module): The projection layer.
        mask_token (nn.Parameter): The mask token.
        cls_token (nn.Parameter): The cls token.

    Example:
        >>> module = VisionEmbedding(224, 16, 3, 768)
        >>> x = torch.randn(2, 3, 224, 224)
        >>> y = module(x)
        >>> y.shape
        torch.Size([2, 197, 768])

    """

    def __init__(
        self,
        img_size=224,
        patch_size=16,
        in_chans=3,
        embed_dim=768,
        contain_mask_token=False,
        prepend_cls_token=False,
    ):
        super().__init__()
        img_size = (img_size, img_size)
        patch_size = (patch_size, patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (
            img_size[0] // patch_size[0]
        )
        self.patch_shape = (
            img_size[0] // patch_size[0],
            img_size[1] // patch_size[1],
        )
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(
            in_chans, embed_dim, kernel_size=patch_size, stride=patch_size
        )

        if contain_mask_token:
            self.mask_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        else:
            self.mask_token = None

        if prepend_cls_token:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        else:
            self.cls_token = None

    def num_position_embeddings(self):
        """num_position_embeddings"""
        if self.cls_token is None:
            return self.num_patches
        else:
            return self.num_patches + 1

    def forward(self, x, masked_position=None, **kwargs):
        """forward"""
        B, C, H, W = x.shape
        assert H == self.img_size[0] and W == self.img_size[1], (
            f"Input image size ({H}*{W}) doesn't match model"
            f" ({self.img_size[0]}*{self.img_size[1]})."
        )
        x = self.proj(x).flatten(2).transpose(1, 2)

        batch_size, seq_len, _ = x.size()

        if masked_position is not None:
            assert self.mask_token is not None
            mask_token = self.mask_token.expand(batch_size, seq_len, -1)
            w = masked_position.unsqueeze(-1).type_as(mask_token)
            x = x * (1 - w) + mask_token * w

        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(
                batch_size, -1, -1
            )  # stole cls_tokens impl from Phil Wang, thanks
            x = torch.cat((cls_tokens, x), dim=1)

        return x



from transformers import AutoTokenizer


class LanguageTokenizerGPTX:
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(
            "EleutherAI/gpt-neox-20b",
            eos_token="<eos>",
            pad_token="<pad>",
            extra_ids=0,
            model_max_length=8192,
        )

    def tokenize_texts(self, texts):
        return self.tokenizer(
            texts, return_tensors="pt", padding=True, truncation=True
        ).input_ids

    def decode(self, texts):
        return self.tokenizer.decode(texts)

    def __len__(self):
        num_tokens = len(self.tokenizer)
        return num_tokens



# from lucidrain


import torch
import torch.nn.functional as F
from torch import nn, Tensor
from torch.nn import Module, ModuleList

from collections import namedtuple

from vector_quantize_pytorch.lookup_free_quantization import LFQ

from einops import rearrange, repeat, reduce, pack, unpack
from einops.layers.torch import Rearrange

from beartype import beartype
from beartype.typing import Union, Tuple, Optional

# helper


def exists(v):
    return v is not None


def default(v, d):
    return v if exists(v) else d


def identity(t):
    return t


def divisible_by(num, den):
    return (num % den) == 0


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


def is_odd(n):
    return not divisible_by(n, 2)


def cast_tuple(t, length=1):
    return t if isinstance(t, tuple) else ((t,) * length)


# helper classes


def Sequential(*modules):
    modules = [*filter(exists, modules)]

    if len(modules) == 0:
        return nn.Identity()

    return nn.Sequential(*modules)


class Residual(Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(x, **kwargs) + x


# adaptive conv from Karras et al. Stylegan2
# for conditioning on latents


class AdaptiveConv3DMod(Module):
    @beartype
    def __init__(
        self,
        dim,
        *,
        spatial_kernel,
        time_kernel,
        dim_out=None,
        demod=True,
        eps=1e-8,
    ):
        super().__init__()
        dim_out = default(dim_out, dim)

        self.eps = eps

        assert is_odd(spatial_kernel) and is_odd(time_kernel)

        self.spatial_kernel = spatial_kernel
        self.time_kernel = time_kernel

        self.padding = (
            *((spatial_kernel // 2,) * 4),
            *((time_kernel // 2,) * 2),
        )
        self.weights = nn.Parameter(
            torch.randn(
                (dim_out, dim, time_kernel, spatial_kernel, spatial_kernel)
            )
        )

        self.demod = demod

        nn.init.kaiming_normal_(
            self.weights, a=0, mode="fan_in", nonlinearity="selu"
        )

    def forward(self, fmap, mod: Optional[Tensor] = None):
        """
        notation

        b - batch
        n - convs
        o - output
        i - input
        k - kernel
        """

        b = fmap.shape[0]

        # prepare weights for modulation

        weights = self.weights

        # do the modulation, demodulation, as done in stylegan2

        mod = rearrange(mod, "b i -> b 1 i 1 1 1")

        weights = weights * (mod + 1)

        if self.demod:
            inv_norm = (
                reduce(weights**2, "b o i k0 k1 k2 -> b o 1 1 1 1", "sum")
                .clamp(min=self.eps)
                .rsqrt()
            )
            weights = weights * inv_norm

        fmap = rearrange(fmap, "b c t h w -> 1 (b c) t h w")

        weights = rearrange(weights, "b o ... -> (b o) ...")

        fmap = F.pad(fmap, self.padding)
        fmap = F.conv3d(fmap, weights, groups=b)

        return rearrange(fmap, "1 (b o) ... -> b o ...", b=b)


# strided conv downsamples


class SpatialDownsample2x(Module):
    def __init__(self, dim, dim_out=None, kernel_size=3):
        super().__init__()
        dim_out = default(dim_out, dim)
        self.conv = nn.Conv2d(
            dim, dim_out, kernel_size, stride=2, padding=kernel_size // 2
        )

    def forward(self, x):
        x = rearrange(x, "b c t h w -> b t c h w")
        x, ps = pack_one(x, "* c h w")

        out = self.conv(x)

        out = unpack_one(out, ps, "* c h w")
        out = rearrange(out, "b t c h w -> b c t h w")
        return out


class TimeDownsample2x(Module):
    def __init__(self, dim, dim_out=None, kernel_size=3):
        super().__init__()
        dim_out = default(dim_out, dim)
        self.conv = nn.Conv1d(
            dim, dim_out, kernel_size, stride=2, padding=kernel_size // 2
        )

    def forward(self, x):
        x = rearrange(x, "b c t h w -> b h w c t")
        x, ps = pack_one(x, "* c t")

        out = self.conv(x)

        out = unpack_one(out, ps, "* c t")
        out = rearrange(out, "b h w c t -> b c t h w")
        return out


# depth to space upsamples


class SpatialUpsample2x(Module):
    def __init__(self, dim, dim_out=None):
        super().__init__()
        dim_out = default(dim_out, dim)
        conv = nn.Conv2d(dim, dim_out * 4, 1)

        self.net = nn.Sequential(
            conv,
            nn.SiLU(),
            Rearrange("b (c p1 p2) h w -> b c (h p1) (w p2)", p1=2, p2=2),
        )

        self.init_conv_(conv)

    def init_conv_(self, conv):
        o, i, h, w = conv.weight.shape
        conv_weight = torch.empty(o // 4, i, h, w)
        nn.init.kaiming_uniform_(conv_weight)
        conv_weight = repeat(conv_weight, "o ... -> (o 4) ...")

        conv.weight.data.copy_(conv_weight)
        nn.init.zeros_(conv.bias.data)

    def forward(self, x):
        x = rearrange(x, "b c t h w -> b t c h w")
        x, ps = pack_one(x, "* c h w")

        out = self.net(x)

        out = unpack_one(out, ps, "* c h w")
        out = rearrange(out, "b t c h w -> b c t h w")
        return out


class TimeUpsample2x(Module):
    def __init__(self, dim, dim_out=None):
        super().__init__()
        dim_out = default(dim_out, dim)
        conv = nn.Conv1d(dim, dim_out * 2, 1)

        self.net = nn.Sequential(
            conv, nn.SiLU(), Rearrange("b (c p) t -> b c (t p)", p=2)
        )

        self.init_conv_(conv)

    def init_conv_(self, conv):
        o, i, t = conv.weight.shape
        conv_weight = torch.empty(o // 2, i, t)
        nn.init.kaiming_uniform_(conv_weight)
        conv_weight = repeat(conv_weight, "o ... -> (o 2) ...")

        conv.weight.data.copy_(conv_weight)
        nn.init.zeros_(conv.bias.data)

    def forward(self, x):
        x = rearrange(x, "b c t h w -> b h w c t")
        x, ps = pack_one(x, "* c t")

        out = self.net(x)

        out = unpack_one(out, ps, "* c t")
        out = rearrange(out, "b h w c t -> b c t h w")
        return out


# autoencoder - only best variant here offered, with causal conv 3d


class CausalConv3d(Module):
    @beartype
    def __init__(
        self,
        chan_in,
        chan_out,
        kernel_size: Union[int, Tuple[int, int, int]],
        pad_mode="reflect",
        **kwargs,
    ):
        super().__init__()
        kernel_size = cast_tuple(kernel_size, 3)

        time_kernel_size, height_kernel_size, width_kernel_size = kernel_size

        assert is_odd(height_kernel_size) and is_odd(width_kernel_size)

        dilation = kwargs.pop("dilation", 1)
        stride = kwargs.pop("stride", 1)

        self.pad_mode = pad_mode
        time_pad = dilation * (time_kernel_size - 1) + (1 - stride)
        height_pad = height_kernel_size // 2
        width_pad = width_kernel_size // 2

        self.time_pad = time_pad
        self.time_causal_padding = (
            width_pad,
            width_pad,
            height_pad,
            height_pad,
            time_pad,
            0,
        )

        stride = (stride, 1, 1)
        dilation = (dilation, 1, 1)
        self.conv = nn.Conv3d(
            chan_in,
            chan_out,
            kernel_size,
            stride=stride,
            dilation=dilation,
            **kwargs,
        )

    def forward(self, x):
        pad_mode = self.pad_mode if self.time_pad < x.shape[2] else "constant"

        x = F.pad(x, self.time_causal_padding, mode=pad_mode)
        return self.conv(x)


@beartype
def ResidualUnit(
    dim,
    kernel_size: Union[int, Tuple[int, int, int]],
    pad_mode: str = "reflect",
):
    return Residual(
        Sequential(
            CausalConv3d(dim, dim, kernel_size, pad_mode=pad_mode),
            nn.ELU(),
            CausalConv3d(dim, dim, 1, pad_mode=pad_mode),
            nn.ELU(),
        )
    )


class CausalConvTranspose3d(Module):
    def __init__(
        self,
        chan_in,
        chan_out,
        kernel_size: Union[int, Tuple[int, int, int]],
        *,
        time_stride,
        **kwargs,
    ):
        super().__init__()
        kernel_size = cast_tuple(kernel_size, 3)

        time_kernel_size, height_kernel_size, width_kernel_size = kernel_size

        assert is_odd(height_kernel_size) and is_odd(width_kernel_size)

        self.upsample_factor = time_stride

        height_pad = height_kernel_size // 2
        width_pad = width_kernel_size // 2

        stride = (time_stride, 1, 1)
        padding = (0, height_pad, width_pad)

        self.conv = nn.ConvTranspose3d(
            chan_in, chan_out, kernel_size, stride, padding=padding, **kwargs
        )

    def forward(self, x):
        assert x.ndim == 5
        t = x.shape[2]

        out = self.conv(x)

        out = out[..., : (t * self.upsample_factor), :, :]
        return out


# video tokenizer class

LossBreakdown = namedtuple("LossBreakdown", ["recon_loss", "lfq_entropy_loss"])


class VideoTokenizer(Module):
    """
    Video Tokenizer class:

    - encodes video into tokens
    - decodes tokens back into video
    - quantizes tokens with lookup-free quantization

    Args:
        layers: tuple of tuples of layer types and dimensions
        residual_conv_kernel_size: kernel size for residual convolutions
        num_codebooks: number of codebooks to use
        codebook_size: size of each codebook
        channels: number of channels in video
        init_dim: initial dimension
        input_conv_kernel_size: kernel size for input convolution
        output_conv_kernel_size: kernel size for output convolution
        pad_mode: padding mode for convolutions
        lfq_entropy_loss_weight: weight for entropy loss
        lfq_diversity_gamma: gamma for diversity loss

    Returns:
        recon_video: reconstructed video
        total_loss: total loss
        loss_breakdown: namedtuple of recon_loss and lfq_entropy_loss

    Usage:
        video_tokenizer = VideoTokenizer()
        video_tokenizer(video, video_or_images, return_loss=True)


    """

    @beartype
    def __init__(
        self,
        layers: Tuple[Tuple[str, int], ...] = (
            ("residual", 64),
            ("residual", 64),
            ("residual", 64),
        ),
        residual_conv_kernel_size=3,
        num_codebooks=1,
        codebook_size=8192,
        channels=3,
        init_dim=64,
        input_conv_kernel_size: Tuple[int, int, int] = (7, 7, 7),
        output_conv_kernel_size: Tuple[int, int, int] = (3, 3, 3),
        pad_mode: str = "reflect",
        lfq_entropy_loss_weight=0.1,
        lfq_diversity_gamma=1.0,
    ):
        super().__init__()

        # encoder

        self.conv_in = CausalConv3d(
            channels, init_dim, input_conv_kernel_size, pad_mode=pad_mode
        )

        self.encoder_layers = ModuleList([])
        self.decoder_layers = ModuleList([])

        self.conv_out = CausalConv3d(
            init_dim, channels, output_conv_kernel_size, pad_mode=pad_mode
        )

        dim = init_dim
        time_downsample_factor = 1

        for layer_type, dim_out in layers:
            if layer_type == "residual":
                assert dim == dim_out

                encoder_layer = ResidualUnit(dim, residual_conv_kernel_size)
                decoder_layer = ResidualUnit(dim, residual_conv_kernel_size)

            elif layer_type == "compress_space":
                encoder_layer = SpatialDownsample2x(dim, dim_out)
                decoder_layer = SpatialUpsample2x(dim_out, dim)

            elif layer_type == "compress_time":
                encoder_layer = TimeDownsample2x(dim, dim_out)
                decoder_layer = TimeUpsample2x(dim_out, dim)

                time_downsample_factor *= 2
            else:
                raise ValueError(f"unknown layer type {layer_type}")

            self.encoder_layers.append(encoder_layer)
            self.decoder_layers.insert(0, decoder_layer)

            dim = dim_out

        self.time_padding = time_downsample_factor - 1

        # lookup free quantizer(s) - multiple codebooks is possible
        # each codebook will get its own entropy regularization

        self.quantizers = LFQ(
            dim=dim,
            codebook_size=codebook_size,
            num_codebooks=num_codebooks,
            entropy_loss_weight=lfq_entropy_loss_weight,
            diversity_gamma=lfq_diversity_gamma,
        )

    @beartype
    def encode(self, video: Tensor, quantize=False):
        """Encode video into tokens"""
        x = self.conv_in(video)

        for fn in self.encoder_layers:
            x = fn(x)

        maybe_quantize = identity if not quantize else self.quantizers

        return maybe_quantize(x)

    @beartype
    def decode(self, codes: Tensor):
        """Decode tokens into video"""
        x = codes

        for fn in self.decoder_layers:
            x = fn(x)

        return self.conv_out(x)

    @beartype
    def forward(
        self,
        video,
        video_or_images: Tensor,
        return_loss=False,
        return_codes=False,
    ):
        """
        Forward pass for video tokenizer

        Args:
            video: video tensor
            video_or_images: video or images tensor
            return_loss: whether to return loss
            return_codes: whether to return codes

        Returns:
            recon_video: reconstructed video
            total_loss: total loss
            loss_breakdown: namedtuple of recon_loss and lfq_entropy_loss
            codes: codes tensor

        """
        assert not (return_loss and return_codes)
        assert video_or_images.ndim in {4, 5}

        # accept images for image pretraining (curriculum learning from images to video)

        if video_or_images.ndim == 4:
            video = rearrange(video, "b c ... -> b c 1 ...")
        else:
            video = video_or_images

        # pad the time, accounting for total time downsample factor, so that images can be trained independently

        padded_video = F.pad(
            video, (0, 0, 0, 0, self.time_padding, 0), value=0.0
        )

        # encoder

        x = self.encode(padded_video)

        # lookup free quantization

        quantized, codes, aux_losses = self.quantizers(x)

        if return_codes:
            return codes

        # decoder

        padded_recon_video = self.decode(quantized)

        recon_video = padded_recon_video[:, :, self.time_padding :]

        # reconstruction loss

        if not return_loss:
            return recon_video

        recon_loss = F.mse_loss(video, recon_video)

        total_loss = recon_loss + aux_losses

        return total_loss, LossBreakdown(recon_loss, aux_losses)


# main class

# class MagViT2(Module):
#     def __init__(self):
#         super().__init__()

#     def forward(self, x):
#         return x



from torch import nn


class MLP(nn.Module):
    """
    Multi-Layer Perceptron (MLP) module.

    Args:
        dim_in (int): The dimension of the input tensor.
        dim_out (int): The dimension of the output tensor.
        expansion_factor (float, optional): The expansion factor for the hidden dimension. Default is 2.
        depth (int, optional): The number of hidden layers. Default is 2.
        norm (bool, optional): Whether to apply layer normalization to the hidden layers. Default is False.

    Attributes:
        net (nn.Sequential): The sequential module for the MLP.

    #Usage
    ```
    from zeta.nn import MLP

    mlp = MLP(
        dim_in=256,
        dim_out=10,
        expansion_factor=4.,
        depth=3,
        norm=True
    )

    x = torch.randn(32, 256)

    #apply the mlp
    output = mlp(x)

    #output tensor
    print(output)
    ```

    """

    def __init__(
        self,
        dim_in: int,
        dim_out: int,
        *,
        expansion_factor=2.0,
        depth=2,
        norm=False,
    ):
        super().__init__()
        hidden_dim = int(expansion_factor * dim_out)

        def norm_fn():
            return nn.LayerNorm(hidden_dim) if norm else nn.Identity()

        layers = [
            nn.Sequential(nn.Linear(dim_in, hidden_dim), nn.SiLU(), norm_fn())
        ]

        for _ in range(depth - 1):
            layers.append(
                nn.Sequential(
                    nn.Linear(hidden_dim, hidden_dim), nn.SiLU(), norm_fn()
                )
            )
        layers.append(nn.Linear(hidden_dim, dim_out))
        self.net = nn.Sequential(*layers)

    def forward(self, x):
        """
        Forward pass of the MLP module.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.

        """
        return self.net(x.float())
